#ifndef PEBBLISP_H
#define PEBBLISP_H

#include "object.h"

#ifndef STANDALONE
#include <pebble.h>
#define printd(...) copySlice(NULL, NULL)
#endif

#ifdef DEBUG
#define printd(...) printf(__VA_ARGS__)
#else
#define printd(...) copySlice(NULL, NULL)
#endif

struct Slice {
    const char *text;
    char length;
};

typedef struct Result {
    Object obj;
    struct Slice *slices;
} Result;

struct Environment {
    char **strings;
    Object *objects;
    struct Environment *outer;
    char size;
};

Object eval(const Object *obj, struct Environment *env);
Result parse(struct Slice *slices);
Result readSeq(struct Slice *slices);
Object parseAtom(struct Slice *slice);
void copySlice(char * dest, struct Slice *src);
Object parseEval(const char *input, struct Environment *env);
void eval_forms(Object *destList, const Object *src, struct Environment *env);

Object evalLambdaArgs(const Object *arg_forms);

struct Environment defaultEnv();
void deleteEnv(struct Environment *e);
void addToEnv(struct Environment *env, const char *name, const Object obj);
Object fetchFromEnvironment(const char *name, struct Environment *env);
void printEnv(struct Environment *env);
struct Environment envForLambda(const Object *params, const Object *arg_forms,
                                struct Environment *outer);

Result result(Object obj, struct Slice *slices);

// Slices
void copySlice(char * dest, struct Slice *src);
void debugSlice(struct Slice *s);

#endif
